/*
 *  Copyright (C) 2000 heXoNet Support GmbH, D-66424 Homburg.
 *  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

#include "rfbServer.h"
#include <stdlib.h>
#include <string.h>


namespace rfb {



void computeSubrectDimensions8( unsigned char *, unsigned int &, unsigned int &, unsigned int );
void computeSubrectDimensions16( unsigned char *, unsigned int &, unsigned int &, unsigned int );
void computeSubrectDimensions24( unsigned char *, unsigned int &, unsigned int &, unsigned int );
void computeSubrectDimensions32( unsigned char *, unsigned int &, unsigned int &, unsigned int );



unsigned char* Server::encodeRectangleRRE( unsigned int   _x,
                                           unsigned int   _y,
                                           unsigned int   _width,
                                           unsigned int   _height,
                                           unsigned char* _buffer,
                                           unsigned int&  _size,
                                           unsigned int   _maxSize )
{
  if ( _maxSize < 4 ) return NULL;
  _size = 0;
  bool visited[_width * _height];
  unsigned int i = 0;
  while ( i < _width * _height )
    visited[i++] = false;
  CARD32 *subrectangles = (CARD32 *) _buffer;
  *subrectangles = 0;
  _size += 4;


  Framebuffer *cframebuffer;
  if ( clientPixelFormat ) {
    cframebuffer = new Framebuffer;
    cframebuffer->pixelFormat = (*clientPixelFormat);
    cframebuffer->bytesPerLine = _width * (clientPixelFormat->bits_per_pixel >> 3 );
    cframebuffer->data = (unsigned char*) malloc( cframebuffer->bytesPerLine * _height );
    unsigned int cy;
    unsigned char *src = framebuffer->data
                         + _x * (framebuffer->pixelFormat.bits_per_pixel >> 3 )
                         + _y * framebuffer->bytesPerLine;
    for ( cy = 0; cy < _height; cy++ ) {
      encodePixels( cframebuffer->data + cy * cframebuffer->bytesPerLine, src, _width );
      src += framebuffer->bytesPerLine;
    }
    cframebuffer->data -= ( _x * (cframebuffer->pixelFormat.bits_per_pixel >> 3 )
                            + _y * cframebuffer->bytesPerLine );
  } else {
    cframebuffer = framebuffer;
  }
  
  switch ( cframebuffer->pixelFormat.bits_per_pixel ) {

    case 8: {
      if ( _maxSize < _size + 1 ) {
        if ( clientPixelFormat ) {
          cframebuffer->data += ( _x * (cframebuffer->pixelFormat.bits_per_pixel >> 3 )
                                  + _y * cframebuffer->bytesPerLine );
	  free( cframebuffer->data );
	  free( cframebuffer );
	}
        return NULL;
      }
      *((CARD8 *) (_buffer + _size)) = 0;
      _size += 1;
      unsigned int x, y, w, h;
      unsigned char *src;
      for ( y = 0; y < _height; y++ ) {
        x = 0;
        while ( x < _width ) {
          if ( !visited[y*_width + x] ) {
            src = cframebuffer->data + ((_x + x) * 1 + (_y + y) * cframebuffer->bytesPerLine);
            if ( src[0] ) {  // black background
              if ( _maxSize < _size + 9 ) {
                if ( clientPixelFormat ) {
                  cframebuffer->data += ( _x * (cframebuffer->pixelFormat.bits_per_pixel >> 3 )
                                         + _y * cframebuffer->bytesPerLine );
	          free( cframebuffer->data );
	          free( cframebuffer );
     	        }
                return NULL;
              }
              (*subrectangles) = (*subrectangles) + 1;
              w = _width - x;
              h = _height - y;
              computeSubrectDimensions8( src, w, h, cframebuffer->bytesPerLine );
              unsigned int x1, y1;
              for ( y1 = y; y1 < h; y1++ )
                for ( x1 = x; x1 < w; x1++ )
                  visited[y1 * _width + x1] = true;
              CARD16 x_position = x;
              CARD16 y_position = y;
              CARD16 width = w;
              CARD16 height = h;
              memcpy( _buffer + _size + 0, src,         1 );
              memcpy( _buffer + _size + 1, &x_position, 2 );
              memcpy( _buffer + _size + 3, &y_position, 2 );
              memcpy( _buffer + _size + 5, &width,      2 );
              memcpy( _buffer + _size + 7, &height,     2 );
              _size += 9;
              x += w;
            }
            else
              x++;
          }
          else
            x++;
        }
      }
    } break;
    case 16: {
      if ( _maxSize < _size + 2 ) {
        if ( clientPixelFormat ) {
          cframebuffer->data += ( _x * (cframebuffer->pixelFormat.bits_per_pixel >> 3 )
                                  + _y * cframebuffer->bytesPerLine );
	  free( cframebuffer->data );
	  free( cframebuffer );
	}
        return NULL;
      }
      *((CARD16 *) (_buffer + _size)) = 0;
      _size += 2;
      unsigned int x, y, w, h;
      unsigned char *src;
      for ( y = 0; y < _height; y++ ) {
        x = 0;
        while ( x < _width ) {
          if ( !visited[y*_width + x] ) {
            src = cframebuffer->data + ((_x + x) * 2 + (_y + y) * cframebuffer->bytesPerLine);
            if ( src[0] || src[1] ) {  // black background
              if ( _maxSize < _size + 10 ) {
                if ( clientPixelFormat ) {
                  cframebuffer->data += ( _x * (cframebuffer->pixelFormat.bits_per_pixel >> 3 )
                                         + _y * cframebuffer->bytesPerLine );
	          free( cframebuffer->data );
	          free( cframebuffer );
     	        }
                return NULL;
              }
              (*subrectangles) = (*subrectangles) + 1;
              w = _width - x;
              h = _height - y;
              computeSubrectDimensions16( src, w, h, cframebuffer->bytesPerLine );
              unsigned int x1, y1;
              for ( y1 = y; y1 < h; y1++ )
                for ( x1 = x; x1 < w; x1++ )
                  visited[y1 * _width + x1] = true;
              CARD16 x_position = x;
              CARD16 y_position = y;
              CARD16 width = w;
              CARD16 height = h;
              memcpy( _buffer + _size + 0, src,         2 );
              memcpy( _buffer + _size + 2, &x_position, 2 );
              memcpy( _buffer + _size + 4, &y_position, 2 );
              memcpy( _buffer + _size + 6, &width,      2 );
              memcpy( _buffer + _size + 8, &height,     2 );
              _size += 10;
              x += w;
            }
            else
              x++;
          }
          else
            x++;
        }
      }
    } break;
    case 32: {
      if ( _maxSize < _size + 4 ) {
        if ( clientPixelFormat ) {
          cframebuffer->data += ( _x * (cframebuffer->pixelFormat.bits_per_pixel >> 3 )
                                  + _y * cframebuffer->bytesPerLine );
	  free( cframebuffer->data );
	  free( cframebuffer );
	}
        return NULL;
      }
      *((CARD32 *) (_buffer + _size)) = 0;
      _size += 4;
      unsigned int x, y, w, h;
      unsigned char *src;
      for ( y = 0; y < _height; y++ ) {
        x = 0;
        while ( x < _width ) {
          if ( !visited[y*_width + x] ) {
            src = cframebuffer->data + ((_x + x) * 4 + (_y + y) * cframebuffer->bytesPerLine);
            if ( src[0] || src[1] || src[2] || src[3] ) {  // black background
              if ( _maxSize < _size + 12 ) {
                if ( clientPixelFormat ) {
                  cframebuffer->data += ( _x * (cframebuffer->pixelFormat.bits_per_pixel >> 3 )
                                         + _y * cframebuffer->bytesPerLine );
	          free( cframebuffer->data );
	          free( cframebuffer );
     	        }
                return NULL;
              }
              (*subrectangles) = (*subrectangles) + 1;
              w = _width - x;
              h = _height - y;
              computeSubrectDimensions32( src, w, h, cframebuffer->bytesPerLine );
              unsigned int x1, y1;
              for ( y1 = y; y1 < h; y1++ )
                for ( x1 = x; x1 < w; x1++ )
                  visited[y1 * _width + x1] = true;
              CARD16 x_position = x;
              CARD16 y_position = y;
              CARD16 width = w;
              CARD16 height = h;
              memcpy( _buffer + _size + 0, src,         4 );
              memcpy( _buffer + _size + 4, &x_position, 2 );
              memcpy( _buffer + _size + 6, &y_position, 2 );
              memcpy( _buffer + _size + 8, &width,      2 );
              memcpy( _buffer + _size + 10, &height,     2 );
              _size += 12;
              x += w;
            }
            else
              x++;
          }
          else
            x++;
        }
      }
    } break;
    
  }
  if ( clientPixelFormat ) {
    cframebuffer->data += ( _x * (cframebuffer->pixelFormat.bits_per_pixel >> 3 )
                            + _y * cframebuffer->bytesPerLine );
    free( cframebuffer->data );
    free( cframebuffer );
  }
  return _buffer;
}


} // namespace rfb
